# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store celery tasks.*

:details: lara_django_store celery tasks.
"""

# from config import celery_app

# @celery_app.task()
