# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store urls.*

:details: lara_django_store urls module.
         - add app specific urls here
"""

# !! this sets the apps namespace to be used in the template
app_name = "lara_django_store"
