# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store admin.*

:details: lara_django_store admin module admin backend configuration.
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev admin_generator lara_django_store >> admin.py" to update this file
________________________________________________________________________
"""

# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import HazardClass, SafetyInformation


@admin.register(HazardClass)
class HazardClassAdmin(admin.ModelAdmin):
    """Admin interface for HazardClass model."""

    list_display = (
        "name",
        "iri",
        "description",
        "hazardclass_id",
    )


@admin.register(SafetyInformation)
class SafetyInformationAdmin(admin.ModelAdmin):
    """Admin interface for SafetyInformation model."""
