"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_store Model Context Protocol (MCP) handler *

:details: lara_django_store Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from mcp_server import ModelQueryToolset

from .models import HazardClass, SafetyInformation


class HazardClassQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `HazardClass`."""

    model = HazardClass


class SafetyInformationQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `SafetyInformation`."""

    model = SafetyInformation
