# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store forms.*

:details: lara_django_store admin module admin backend configuration.
"""
