# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store app.*

:details: lara_django_store app configuration.
         This provides a generic django app configuration mechanism.
         For more details see:
         https://docs.djangoproject.com/en/4.0/ref/applications/
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from django.apps import AppConfig


class LaraStoreConfig(AppConfig):
    """Configuration for the lara_django_store Django app."""

    name = "lara_django_store"
    # url_path = 'store'
    # enter a verbose name for your app: lara_django_store here - this will be used in the admin interface
    verbose_name = "LARA-django Store"
    # lara_app_icon = 'lara_django_store_icon.svg'  # this will be used to display an icon, e.g. in the main LARA menu.
